/***********************************************************************
 *
 * PROJECT: ClockupDA
 *
 * FILE: Starter.c
 *
 * DESCRIPTION:	Clock up/down DA for Palm
 *
 * AUTHOR: Hiroyuki Okamoto
 * mailto:HiroyukiOkamoto@yahoo.co.jp
 * http://hokamoto.tripod.com/
 * 
 * DISTRIBUTION AND RESTRICTIONS ON USAGE:
 *  1) Please give this source code away to your friends at no charge.
 *  2) Only use this tool on a copy of a file -- do not use an original. This
 *     is just common sense.
 *  3) This source code or a compiled version may be bundled with commercial
 *     software as long as the author is notified beforehand. The author's name
 *     should also be mentioned in the credits.
 *  4) Feel free to use any of the algorithms for your own work. Many of them are
 *     being used in other tools I have written.
 *  5) The most current version can be obtained by requesting a copy directly
 *    from me.
 *	  
 **********************************************************************/

#include <Pilot.h>
#include <FeatureMgr.h>
#include <NetMgr.h>
#include <Preferences.h>
#include <DateTime.h>
#include "StarterRsc.h"

#define maxClock		42
#define minClock		10


#define ChrIsHardKey(c)		((((c) >= hardKeyMin) && ((c) <= hardKeyMax)) || ((c) == calcChr))
#define appFileCreator			'CUda'
#define appVersionNum              0x01
#define appPrefID                  0x00
#define appPrefVersionNum          0x02

#define	PLLFREQ		0xfffff202			// PLL Frequency Control Register
#define	DRAMCTRL	0xfffffc02			// DRAM Control Register

typedef struct 
	{
	UInt16 defaultClock;
	UInt16 currentClock;
	Boolean dramNoWait;
	} StarterPreferenceType;

static void InitMainForm();
static void SavePrefs();
static void EventLoop();
static Boolean HandleEvent(EventPtr);
static VoidPtr GetObjectPtr(Word);
void Display();

void start()
{
    FormPtr formP;

    formP= FrmInitForm(MainForm);
    FrmSetActiveForm(formP);
    InitMainForm();
    FrmUpdateForm(MainForm, 0);
    EventLoop();
    FrmEraseForm(formP);
    FrmDeleteForm(formP);
}

static VoidPtr GetObjectPtr(Word id)
{
    FormPtr form;

    form= FrmGetActiveForm();
    return FrmGetObjectPtr(form, FrmGetObjectIndex(form, id));
}

static void EventLoop()
{
    EventType event;
    Boolean handled= false;

    do {
        EvtGetEvent(&event, 0);
        if (SysHandleEvent(&event) == false &&
            FrmHandleEvent(FrmGetActiveForm(), &event) == false) {
            handled= HandleEvent(&event);
        }
    } while (handled == false);

    SavePrefs();   
}

/* check if this processor is DRAGONBALL EZ or not */
static Boolean IsProcessorEZ()
{

// If we're not using a new include file, define things appropriately:
#ifndef sysFtrNumProcessorID
#define sysFtrNumProcessorID 2
// Product id
// 0xMMMMRRRR, where MMMM is the processor model and RRRR is the revision.
#define sysFtrNumProcessorMask 0xFFFF0000	// Mask to obtain processor model
#define sysFtrNumProcessor328 0x00010000	// Motorola 68328 (Dragonball)
#define sysFtrNumProcessorEZ 0x00020000		// Motorola 68EZ328 (Dragonball EZ)
#endif

	DWord   id, chip;
	Err     err;

	err = FtrGet(sysFtrCreator, sysFtrNumProcessorID, &id);
	if (!err)
		{ 
		chip     = id & sysFtrNumProcessorMask;
		return (chip >= sysFtrNumProcessorEZ) ? true : false;
		}
	return false;
}


/***********************************************************************
 *
 * FUNCTION:    SetFreq
 *
 * DESCRIPTION: This routine changes clock(VCO frequency).
 *              it sets QC & PC values in PLL Frequency Select Register(PLLSR)
 *              clock(Hz) = 32 * (14 * (p + 1) + Q + 1) 
 *		see MC68EZ328 User's manual.
 *
 * PARAMETERS:  newFreq     - new QP values in PLLSR
 *
 * RETURNED:    old QP value
 ***********************************************************************/
UInt16 asm SetFreq(UInt16 newFreq)
{
	move.w SR,-(SP)
	link a6,#-2
	
	move.w PLLFREQ,-2(a6)		// save current frequency P&Q value
SYNC1:
	btst #7,PLLFREQ			// synchronize to CLK32 high level
	beq.s SYNC1			// CLK32 is still not high, go back
SYNC2:
	btst #7,PLLFREQ			// synchronize to CLK32 low level
	bne.s SYNC2			// CLK32 is still not low, go back

	move.w 10(a6),PLLFREQ		// load the new frequency P&Q value

SYNC3:
	btst #7,PLLFREQ			// synchronize to CLK32 high level
	beq.s SYNC3			// CLK32 is still not high, go back
SYNC4:
	btst #7,PLLFREQ			// synchronize to CLK32 low level
	bne.s SYNC4			// CLK32 is still not low, go back

	// set previous frequency P&Q value
	move.w -2(a6),d0
	andi.w #0x7fff,d0
	unlk a6
	rtr
}

/* Get current PLLSR value */
static UInt16 GetFreq()
{
	volatile UInt16 *p;
	p = (volatile UInt16 *)PLLFREQ;
	return (0x7fff & *p);
}

/* converts clock value to PLLSR value */
static UInt16 FreqToClock(UInt16 f)
{
	UInt16 p, q;

	/* these are default freq values in Palm */
	if (!f || f == 0x0123)
		return 16;	// 16MHz
	else if (f == 0x012b)
		return 20;	// 20MHz
	else if (f == 0x0347)
		return 33;	// 33MHz

	p = f & 0x00ff;
	q = 1;
	f = 32 * (14 * (p + 1) + q + 1);
	return (f/1000);
}

/* converts PLLSR value to clock value */
static UInt16 ClockToFreq(UInt16 c)
{
	UInt16 f;

	/* these are default freq values in Palm */
	if (c == 16)		// 16MHz
		return 0x0123;
	else if (c == 20)
		return 0x012b;	// 20MHz
	else if (c == 33)
		return 0x0347;	// 33MHz

	for (f = 0x15; f < 0x60; f++)		// 10 -> 40 MHz
		if (FreqToClock(f | 0x0100) == c)
			return (f | 0x0100);
	return 0x0123;		// something wrong, just returns 16MHz
}


/***********************************************************************
 *
 * FUNCTION:    GetDramNoWait
 *
 * DESCRIPTION: This function check DRAM wait-states(WSx) 
                in DRAM Control Register(DRAMC)
 *		see MC68EZ328 User's manual.
 *
 * RETURNED:    true if No wait-states. otherwise, false
 ***********************************************************************/
static Boolean GetDramNoWait()
{
	Word *regDRAMC;
	regDRAMC = (Word *)DRAMCTRL;
	return (*regDRAMC & 0060) ? false : true;
}

static void SetDramNoWait(Boolean on)
{
	Word *regDRAMC;
	regDRAMC = (Word *)DRAMCTRL;
	if (on)
		*regDRAMC &= 0xff9f;		// zero wait-state
	else
		*regDRAMC |= 0x0060;		// three wait-states

}

static void SavePrefs()
{
	StarterPreferenceType prefs;
	UInt16 prefsSize;
	UInt16 currentClock;

	currentClock = FreqToClock(GetFreq());
	if (!currentClock || currentClock > maxClock)
		currentClock = 16;

	prefsSize = sizeof(StarterPreferenceType);
	if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, false) == noPreferenceFound)
		prefs.defaultClock = currentClock;

	prefs.currentClock = currentClock;
	if (IsProcessorEZ())
		prefs.dramNoWait = GetDramNoWait();

	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum, &prefs, sizeof (prefs), false);

}

static void Display()
{
	FormPtr formP;
	UInt16 clock;
	char label[32];

	formP = FrmGetActiveForm();
	clock = FreqToClock(GetFreq());

	// Display Clock
	StrIToA(label, clock);
	StrCopy(&label[StrLen(label)], " MHz");
	FrmCopyLabel(formP, MainClocksLabel, label);
	FrmHideObject(formP, FrmGetObjectIndex(formP, MainClocksLabel));
	FrmShowObject(formP, FrmGetObjectIndex(formP, MainClocksLabel));


	// Display DRAM
	if (IsProcessorEZ())
		FrmSetControlGroupSelection(formP, 1, GetDramNoWait() ? MainFastPushButton : MainSlowPushButton);
	else
		{
		FrmHideObject(formP, FrmGetObjectIndex(formP, MainFastPushButton));
		FrmHideObject(formP, FrmGetObjectIndex(formP, MainSlowPushButton));
		FrmHideObject(formP, FrmGetObjectIndex(formP, MainDramLabel));
		}

}


static void InitMainForm()
{
	SavePrefs();   
	Display();
}


void ClockUpDown(Boolean isUp)
{
	UInt16 clock;
	clock = FreqToClock(GetFreq());
	if (isUp && clock < maxClock)
		{
		clock++;
		SetFreq(ClockToFreq(clock));
		Display();
		}
	else if (!isUp && clock > minClock)
		{
		clock--;
		SetFreq(ClockToFreq(clock));
		Display();
		}
}

static Boolean HandleEvent(EventPtr event)
{
    Boolean handled = false;
    StarterPreferenceType prefs;
    UInt16 prefsSize;
 
    switch (event->eType) {

	 
       case ctlEnterEvent:
	    if (event->data.ctlEnter.controlID == MainUpRepeating)
		{
		ClockUpDown(true);
		}
	    else if (event->data.ctlEnter.controlID == MainDownRepeating)		
		{
		ClockUpDown(false);
		}
	    break;

       case ctlRepeatEvent:
	    if (event->data.ctlRepeat.controlID == MainUpRepeating)
		{
		ClockUpDown(true);
		}
	    else if (event->data.ctlRepeat.controlID == MainDownRepeating)		
		{
		ClockUpDown(false);
		}
	    break;

       case ctlSelectEvent:
	    if (event->data.ctlSelect.controlID == MainDefaultButton)
		{
		if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, false) != noPreferenceFound)
			SetFreq(ClockToFreq(prefs.defaultClock));
		Display();
		}
	    else if (event->data.ctlSelect.controlID == MainFastPushButton)
		{
		if (IsProcessorEZ())
			SetDramNoWait(true);
		}
	    else if (event->data.ctlSelect.controlID == MainSlowPushButton)
		{
		if (IsProcessorEZ())
			SetDramNoWait(false);
		}
	    break;

        case appStopEvent:
            EvtAddEventToQueue(event);
            handled= true;
            break;

        case keyDownEvent:
	    if (event->data.keyDown.chr == chrPageUp)
	        {
		ClockUpDown(true);
		}
	    else if (event->data.keyDown.chr == chrPageDown)	
	        {
		ClockUpDown(false);
		}
            else if (ChrIsHardKey(event->data.keyDown.chr))
		{
                EvtAddEventToQueue(event);
                handled= true;
		}
            break;

        case penDownEvent:
	    {
            SWord x,y;
            RectangleType r;

            x= event->screenX;
            y= event->screenY;
            WinWindowToDisplayPt(&x, &y);
            FrmGetFormBounds(FrmGetActiveForm(), &r);
            if (x < r.topLeft.x || x > (r.topLeft.x+r.extent.x) || y < r.topLeft.y || y > (r.topLeft.y+r.extent.y))
		{
                handled= true;
		}
            break;

        }

        default:
            break;
    }

    return handled;
}

